import java.io.*;
import lejos.nxt.remote.*;
import lejos.pc.comm.*;
import java.util.Arrays;
import java.net.URLDecoder;
import lejos.pc.tools.*;
import java.util.Vector;
import java.util.Iterator;


public class NXTMessenger extends Messenger {
	
	private static final String CONNECTION_STATUS_QUERY = "Connection status?";
	private static final String CONNECTION_STATUS_RESPONSE = "Connection status: ";
	private static final String GENERIC_ERROR = "Generic Error: ";
	
	private static final String NXT_COMMAND_PREFIX = "NXT Command: ";
	private static final String COMPILER_STATUS_PREFIX = "NXT Compilation Status: ";	
	private static final String COMPILER_ERROR_PREFIX = "NXT Compilation Error: ";	
	private static final String COMPILER_DONE = "NXT Compilation Finished.";	
	private static final String FLASH_STATUS_PREFIX = "NXT FirmwareFlasher Status: ";
	private static final String FLASH_ERROR_PREFIX = "NXT FirmwareFlasher Error: ";
	private static final String FLASH_DONE = "NXT FirmwareFlasher Finished.";
	private static final String NOT_CONNECTED = "NXT is not connected.";
	
	private String nxtname = "";
	private String address = "";
	private int protocols = NXTCommFactory.USB; // | NXTCommFactory.BLUETOOTH;
	private NXTCommand nxtcommand = new NXTCommand();
	private NXTConnector nxtconnector = new NXTConnector();
	
	private boolean connection_lost = true;
	private boolean uploading = false;
	private boolean user_program_running = false;
	
	static String nxjc;    /** command to run nxjc */
	static String nxjlink; /** command to run nxjlink */
	
	static {
		if (System.getProperty("os.name").toLowerCase().contains("win")) {
			nxjc = "nxjc.bat";
			nxjlink = "nxjlink.bat";
		} else {
			nxjc = "nxjc";
			nxjlink = "nxjlink";
		}
	}

	public NXTMessenger(String name, Dispatcher dispatcher, 
						int connectionDelay, int verbosity) {
		super(name, dispatcher, connectionDelay, verbosity);
	}
	
	/** Repeatedly tries to connect to an NXT.
	    Does not return unless a connection is established. */
	public void connect() {
		boolean connected = false;
		while (!connected) {
			try {
				sleep(connectionDelay);		
			} catch (InterruptedException e) {
				// we are unlikely to be interrupted
				throw new RuntimeException(e);
			}
			//if (verbosity > 8) {
				System.out.println("Trying to connect to NXT");
			//}
			
			try {
				connected = nxtconnector.connectTo(nxtname, address, protocols);
			} catch (java.lang.RuntimeException e) {
				System.out.println("RuntimeException occurred while trying to connect to NXT: " + e.getMessage());
				if (e.getMessage().equals("Can't initialize bluetooth support")) {
					narrowcastMsg(GENERIC_ERROR + "Bluetooth is turned off on this computer.  Disabling bluetooth scanning.");
					protocols = NXTCommFactory.USB;
				}
			}
		}
		System.out.println("Connected to NXT");
		nxtcommand.setNXTComm(nxtconnector.getNXTComm());
	}
	
	/** Listens for messages from the NXT,
	    and forwards them to the dispatcher.  Does not exit
	    unless communication is broken. */
	public void receiveMessages() {
		// Right now the NXT does not send any messages to us
		// at this time.  However, unless we know that communication
		// has been lost, this function continues.
		
		lostConnection(false, "Started receivingMessages()");
		System.out.println("NXT started receiving messages");
		while (!connection_lost) {
			try {
				sleep(500);		
			} catch (InterruptedException e) {
				// we are unlikely to be interrupted
				throw new RuntimeException(e);
			}
			if (uploading) continue;
			
			// Broadcast any message received
			try {
				// broadcast message - processes all in queue	
				while(!connection_lost && !uploading) {
					byte[] message = nxtcommand.messageRead((byte) 0, (byte) 0, true);
					
					if (message.length == 0) break;
					
					String msgString = new String(message);
					System.out.println("Broadcasting: " + msgString);
					broadcastMsg(msgString);
				}
				// sensor updates - processes all in queue
				while (!connection_lost && !uploading) {
					byte[] message = nxtcommand.messageRead((byte) 1, (byte) 0, true);
					if (message.length == 0) break;
					String msgString = new String(message);
					System.out.println("sensor update: " + msgString);
					
					//		updateMsg(msgString);
					// Send a standard update message to Scratch
					forwardTo(new ScratchMessage("sensor-update " + msgString), "Scratch Messenger");
					// and a slightly modified one to Enchanting -- each variable name starts with Enc^.
					forwardTo(new ScratchMessage("sensor-update Enc^" + msgString), "Enchanting Messenger");
				}
				
			} catch (IOException e) {
				lostConnection(true, "IOException occurred in receiveMessages(): " + e.getMessage());
			}
		}
		System.out.println("NXT stopped receiving messages");
		lostConnection(true, "Exited receiveMessages()");
	}
	
	/** Returns true is the message is sent successfully (and false if there is no connection) */
	public boolean sendMessage(ScratchMessage msg, Messenger source) {	
	
		System.out.println("Received message: " + msg.getName() + " value = " + msg.getValue());

		// Narrowcast messages are just used for IPC (inter-process communication)
		if (msg.getMessageType() == ScratchMessage.NARROWCAST_MSG)
		{
			return actOnIPCCommand(msg);
		}
		else
		{
			// Message for the NXT
			if (connected && !connection_lost && !uploading) {
				// Send variable updates first -- broadcasts may depend upon them
				if (msg.getMessageType() == ScratchMessage.SENSOR_UPDATE_MSG) {
					try {
						String message = msg.getName() + " " + msg.getValue();
						
						// if the message is from EnchantingFE, it is a variable update
						// send it on port 1
						if (source.name().equals("Enchanting Messenger")) {
							System.out.println("Sending variable update:" + message);
							nxtcommand.messageWrite(message.getBytes(), (byte) 1);
						}
						// if it is from Scratch, it is a remote variable update
						// send it on port 2
						if (source.name().equals("Scratch Messenger")) {
							System.out.println("Sending remote sensor update:" + message);
							nxtcommand.messageWrite(message.getBytes(), (byte) 2);
						}	
					} catch (IOException e) {
						lostConnection(true, "IOException while sending sensor update message: " + e.getMessage());
					}
				} else if (msg.getMessageType() == ScratchMessage.BROADCAST_MSG) {
					try {
						String message = msg.getName();
						System.out.println("Sending broadcast:" + message);
						nxtcommand.messageWrite(message.getBytes(), (byte) 0);
					} catch (IOException e) {
						lostConnection(true, "IOException while sending broadcast message: " + e.getMessage());
					} 				
				}				
			}

			return true;
		}
	}
	
	private boolean actOnIPCCommand(ScratchMessage msg)
	{
		if (msg.getName().startsWith(NXT_COMMAND_PREFIX)) {
			String full_command = msg.getName().substring(NXT_COMMAND_PREFIX.length());
			String[] argv = full_command.split(" ");
			String command = argv[0];	// The command is the first word in the narrowcast message
			
			System.out.println("Received NXT command: " + Arrays.toString(argv));

			if (command.equals("compile-and-upload")) {
				return compileProgram(argv, false);
			} else if (command.equals("compile-and-run")) { 
				return compileProgram(argv, true); 
			} else if (command.equals("flash")) {
				return flashFirmware();
			} else if (command.equals("bluetooth-enable")) {
				protocols = NXTCommFactory.USB | NXTCommFactory.BLUETOOTH;
				System.out.println("Scanning using bluetooth enabled.");
			} else if (command.equals("bluetooth-disable")) {
				protocols = NXTCommFactory.USB;
				System.out.println("Scanning using bluetooth disabled.");
			} else if (command.equals("stop-program")) {
				return stopProgram();
			} else {
				System.out.println("Unknown command for NXTMessenger: " + full_command);
			}
		} 
		else
		{
			// This isn't an NXT-specific message
			if (msg.getName().equals(CONNECTION_STATUS_QUERY)) {
				sendConnectionStatus();
			}
		}
	
		return false; 
	}

	private void lostConnection(boolean lost, String reason) {
	
		if (connection_lost == lost) {
			System.out.println("Ignoring lostConnection(" + String.valueOf(connection_lost) + "): " + reason);
			return;
		}
	
		connection_lost = lost;
		System.out.println("lostConnection(" + String.valueOf(connection_lost) + "): " + reason);
		sendConnectionStatus();
		
		if (lost == true) {
			user_program_running = false;	
			try {
				nxtconnector.close();
			} catch (IOException e) {
				System.out.println("IOException while closing connection: " + e.getMessage());
			}
		}
	}
	
	private void amUploading(boolean up, String reason) {
		uploading = up;
		System.out.println("Uploading = " + String.valueOf(uploading) + " - " + reason);
		sendConnectionStatus();
	}

	private void sendConnectionStatus() {
		boolean is_connected = connected && !connection_lost;
		boolean is_ready = is_connected && !uploading;
		narrowcastMsg(CONNECTION_STATUS_RESPONSE + String.valueOf(is_connected) + " " + String.valueOf(is_ready));	
	}
	
	private boolean amConnected() {
		return connected && !connection_lost;
	}
	
	private boolean compileProgram(String[] argv, boolean run_program) {
		// We are going to compile and upload this code
		
		// Inform the front end
		amUploading(true, "Starting to upload nxj file");
		narrowcastMsg(COMPILER_STATUS_PREFIX + "Compiling.");   			
		
		// Determine the file name.  (It is the only parameter, URL encoded)
		File sourceFile;
		try {
			sourceFile = new File(URLDecoder.decode(argv[1], "UTF-8"));
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}
		if (!sourceFile.exists()) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Compilation Failed. " +
				"Source file does not exist. (" + sourceFile.getAbsolutePath() + ")");
			amUploading(false, "Compiling failed -- source file not found.  Abandoning upload.");
			return true;	// acted on message
		}
		System.out.println("File exists? " + sourceFile.exists());
						
		// Compile the file
		SaferExec se = new SaferExec(120);	// 120 seconds max time to compile
		String result = se.execV(nxjc, sourceFile.getPath());
		//System.out.println("Result: *" + result + "*");
		if (!result.equals("0\n")) { 
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Compilation Failed. " +
				"Error in code generated by Enchanting."); 
			amUploading(false, "Compiling failed.  Abandoning upload.");
			return true;	// acted on message
		}
		
		// Link the file
		narrowcastMsg(COMPILER_STATUS_PREFIX + "Linking.");  			
		String name = sourceFile.getName();
		String classname = name.substring(0, name.length() - 5);	// remove ".java"
		File nxjFile = new File(sourceFile.getParent(), classname + ".nxj");
		result = se.execV(nxjlink, "-cp", sourceFile.getParent(), "-o", nxjFile.getPath(), classname);
		System.out.println("Result: " + result);
		if (!result.equals("0\n")) { 
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Linking Failed. "); 
			amUploading(false, "Linking failed.  Abandoning upload.");
			return true;	// acted on message
		}
	
		if (!amConnected()) {
			// wait ten seconds to see if we reconnect
			narrowcastMsg(COMPILER_STATUS_PREFIX + "Reconnecting.");    			
			
			for (int i = 0; i < 100; ++i) {
				try {
					sleep(100);		
				} catch (InterruptedException e) {
					throw new RuntimeException(e); 	// we are unlikely to be interrupted
				}
				if (amConnected()) break;
			}
			
			if (!amConnected()) {
				narrowcastMsg(COMPILER_ERROR_PREFIX + "Unable to reconnect.");
				System.out.println("Unable to reconnect to NXT");
				amUploading(false, "Unable to reconnect.  Abandoning upload.");
				return true;	// acted on message
			}
		}

/*		// wait a little bit longer for connection to be solid
		try {
			sleep(250);		
		} catch (InterruptedException e) {
			throw new RuntimeException(e); 	// we are unlikely to be interrupted
		}
*/		
		
		// Upload the file and run it
		if (nxjFile.exists() && amConnected()) {
						
			// uploadMedia();
			
			try {
				narrowcastMsg(COMPILER_STATUS_PREFIX + "Uploading.");    			
				String nxtFileName = "Enchanting.nxj";	// must be under 20 characters
				nxtcommand.setVerify(true);
				nxtcommand.uploadFile(nxjFile, nxtFileName);
				nxtcommand.setVerify(false);
				
				if (run_program) {
					narrowcastMsg(COMPILER_STATUS_PREFIX + "Running.");    			
					nxtcommand.startProgram(nxtFileName);
					// Connection is lost when programs run
					lostConnection(true, "Need to reconnect now that new application is starting.");
					user_program_running = true;
				}
				amUploading(false, "nxj file uploaded successfully");
				narrowcastMsg(COMPILER_DONE);
			} catch (IOException e) {
				narrowcastMsg(COMPILER_ERROR_PREFIX + "Uploading Failed. " + e.getMessage());
				amUploading(false, "IOException while uploading nxj file: " + e.getMessage());
				//lostConnection(false);
			}
		} else {
			String error_message;
			if (!amConnected()) {
				error_message = "Uploading Failed.  No NXT is connected.";
			} else {
				error_message = "Upload Failed. NXT executable file does not exist. (" + nxjFile.getAbsolutePath() + ")";
			}
			narrowcastMsg(COMPILER_ERROR_PREFIX + error_message);
			System.out.println(error_message);
		}	
		
		return true;	// we acted on the message
	}

	private boolean flashFirmware() {
		amUploading(true, "Flashing firmware");
		FirmwareFlasher f = new FirmwareFlasher();
		try {
			f.flash();
			System.out.printf("Flashed successfully.  Severing connection.");
			lostConnection(true, "Firmware flashed; must reconnect");
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		amUploading(false, "Done flashing firmware");
		return true;
	}

	private boolean stopProgram() {
		if (user_program_running) {
			System.out.println("Stopping program");
			try {
				nxtcommand.messageWrite("Enchanting-StopProgram".getBytes(), (byte) 0);
				lostConnection(true, "Stopped program; connection lost.");
			} catch (IOException e) {
				lostConnection(true, "IOException while stopping program: " + e.getMessage());
			}
		}
		else {
			System.out.println("Ignoring request to stop program.  No user program running.");
		}
		return true;	// acted on message
	}

	private void uploadMedia() {
		
		// Code created by computer science students at the University of Lethbridge
		// Has an external dependancie on Sox
		// Only partially integrated at the moment
		
		// MEDIA FILE UPLOADING
		broadcastMsg(COMPILER_STATUS_PREFIX + "Uploading Media.");
		
		// Upload each individual file
		Runtime rt = Runtime.getRuntime();
		try {
			
			// If we are in windows...
			if (System.getProperty("os.name").toLowerCase().contains("win")) {
				// Use this path for media
				String path = "..\\FrontEnd\\Media\\Sounds\\Upload\\";
				// Get the media list that was output by enchanting
				BufferedReader mediaList = new BufferedReader(new FileReader( path + "mediaList.dat"));
				String filename;
				Vector<String> uniqueFilenames = new Vector<String>(); // A list of unique filenames to use in uploading to NXT.
				int numberOfFiles = uniqueFilenames.size(); // Keeps track of the number of files in the list.  Used to determine play length of sample.
				double sampleSecondsQuota = 0; // A variable to determine how long each sample can be.
				SaferExec um = new SaferExec();
				
				// While not EOF
				while( mediaList.ready() ) {
					// Read each media filename
					filename = mediaList.readLine();
					// If this is the first filename encountered, add it to the list.
					if(numberOfFiles == 0){
						uniqueFilenames.add(filename);
						numberOfFiles = uniqueFilenames.size();
					}
					// Otherwise, check if it has already been seen.
					else{
						boolean alreadyAdded = false;
						for(int i = 0; i < numberOfFiles; i++){
							if(filename.equals(uniqueFilenames.elementAt(i))){
								alreadyAdded = true;
							}
						}
						// If this is the first encounter, add the filename to the list.
						if(alreadyAdded == false){
							uniqueFilenames.add(filename);
							numberOfFiles = uniqueFilenames.size();
						}
					}
				}
				mediaList.close();
				// Determine the number of seconds that each sample can be, based on available space on NXT being ~130KB.  150KB == 50 seconds at current processing.
				sampleSecondsQuota = 33.569;
				if(uniqueFilenames.size() != 0) {
					sampleSecondsQuota = 33.569 / uniqueFilenames.size();
				}
				
				// Reuse the String variable "filename" for the next stage.
				filename = "";
				// Process the files listed in uniqueFilenames through SoX and upload to robot
				Iterator itr = uniqueFilenames.iterator();
				String nxjhome = System.getenv("NXJ_HOME") + "\\bin\\";
				
				
				while(itr.hasNext()){
					filename = (String)itr.next();
					um.exec("move " + path + filename + " " + path + "input.wav");
					um.exec(".\\sox\\win\\sox " + path + "input.wav -V1 -b 8 -r 4k -e un -c 1 " + path + filename + " gain -n 3 highpass 200 vol 1.67 trim 0 " + sampleSecondsQuota);
					um.exec("del " + path + "input.wav");
					File audFile = new File(path + filename);
					String s1;
					nxtcommand.setVerify(true);
					nxtcommand.uploadFile(audFile, filename);
					nxtcommand.setVerify(false);
					um.exec("del " + path + filename);
				}
				um.exec("del " + path + "mediaList.dat");
				// Also delete any other audio sample files Enchanting exported that were not uploaded.
				um.exec("del " + path + "*.wav");
			}
			// If we in linux...
			else if (System.getProperty("os.name").toLowerCase().contains("nux")) {
				
				// Use this path for media
				String path = "../FrontEnd/Media/Sounds/Upload/";
				// Get the media list that was output by enchanting
				BufferedReader mediaList = new BufferedReader(new FileReader( path + "mediaList.dat"));
				String filename;
				Vector<String> uniqueFilenames = new Vector<String>(); // A list of unique filenames to use in uploading to NXT.
				int numberOfFiles = uniqueFilenames.size(); // Keeps track of the number of files in the list.  Used to determine play length of sample.
				double sampleSecondsQuota = 0; // A variable to determine how long each sample can be.
				SaferExec um = new SaferExec();
				
				// While not EOF
				while( mediaList.ready() ) {
					// Read each media filename
					filename = mediaList.readLine();
					// If this is the first filename encountered, add it to the list.
					if(numberOfFiles == 0){
						uniqueFilenames.add(filename);
						numberOfFiles = uniqueFilenames.size();
					}
					// Otherwise, check if it has already been seen.
					else{
						boolean alreadyAdded = false;
						for(int i = 0; i < numberOfFiles; i++){
							if(filename.equals(uniqueFilenames.elementAt(i))){
								alreadyAdded = true;
							}
						}
						// If this is the first encounter, add the filename to the list.
						if(alreadyAdded == false){
							uniqueFilenames.add(filename);
							numberOfFiles = uniqueFilenames.size();
						}
					}
				}
				mediaList.close();
				// Determine the number of seconds that each sample can be, based on available space on NXT being ~130KB.  150KB == 50 seconds at current processing.
				sampleSecondsQuota = 33.569;
				if(numberOfFiles != 0){
					sampleSecondsQuota = 33.569 / uniqueFilenames.size();
				}
				// Reuse the String variable "filename" for the next stage.
				filename = "";
				// Process the files listed in uniqueFilenames through SoX and upload to robot
				Iterator itr = uniqueFilenames.iterator();
				String nxjhome = System.getenv("NXJ_HOME") + "\\bin\\";
				
				
				while(itr.hasNext()){
					filename = (String)itr.next();
					um.exec("mv " + path + filename + " " + path + "input.wav");
					um.exec("./sox/linux/sox " + path + "input.wav -V1 -b 8 -r 4k -e un -c 1 " + path + filename + " gain -n 3 highpass 200 vol 1.67 trim 0 " + sampleSecondsQuota);
					um.exec("rm -f " + path + "input.wav");
					File audFile = new File(path + filename);
					String s1;
					nxtcommand.setVerify(true);
					nxtcommand.uploadFile(audFile, filename);
					nxtcommand.setVerify(false);
					um.exec("rm -f " + path + filename);
				}
				um.exec("rm -f " + path + "mediaList.dat");
				// Also delete any other audio sample files Enchanting exported that were not uploaded.
				um.exec("rm -f " + path + "*.wav");
			}
			// If we in OSX..
			else if (System.getProperty("os.name").toLowerCase().contains("mac")) {
				
				// Use this path for media
				//String path = "../FrontEnd/Media/Sounds/Upload/";
				String path = "/Media/Sounds/Upload/";
				// Get the media list that was output by enchanting
				BufferedReader mediaList = new BufferedReader(new FileReader( path + "mediaList.dat"));
				String filename;
				Vector<String> uniqueFilenames = new Vector<String>(); // A list of unique filenames to use in uploading to NXT.
				int numberOfFiles = uniqueFilenames.size(); // Keeps track of the number of files in the list.  Used to determine play length of sample.
				double sampleSecondsQuota = 0; // A variable to determine how long each sample can be.
				SaferExec um = new SaferExec();
				
				// While not EOF
				while( mediaList.ready() ) {
					// Read each media filename
					filename = mediaList.readLine();
					// If this is the first filename encountered, add it to the list.
					if(numberOfFiles == 0){
						uniqueFilenames.add(filename);
						numberOfFiles = uniqueFilenames.size();
					}
					// Otherwise, check if it has already been seen.
					else{
						boolean alreadyAdded = false;
						for(int i = 0; i < numberOfFiles; i++){
							if(filename.equals(uniqueFilenames.elementAt(i))){
								alreadyAdded = true;
							}
						}
						// If this is the first encounter, add the filename to the list.
						if(alreadyAdded == false){
							uniqueFilenames.add(filename);
							numberOfFiles = uniqueFilenames.size();
						}
					}
				}
				mediaList.close();
				// Determine the number of seconds that each sample can be, based on available space on NXT being ~130KB.  150KB == 50 seconds at current processing.
				sampleSecondsQuota = 33.569;
				if(uniqueFilenames.size() != 0) {
					sampleSecondsQuota = 33.569 / uniqueFilenames.size();
				}
				
				// Reuse the String variable "filename" for the next stage.
				filename = "";
				// Process the files listed in uniqueFilenames through SoX and upload to robot
				Iterator itr = uniqueFilenames.iterator();
				String nxjhome = System.getenv("NXJ_HOME") + "\\bin\\";
				
				
				while(itr.hasNext()){
					filename = (String)itr.next();
					um.exec("mv " + path + filename + " " + path + "input.wav");
					System.out.println(um.exec("pwd && ls"));
					um.exec("./sox/mac/sox " + path + "input.wav -V1 -b 8 -r 4k -e un -c 1 " + path + filename + " gain -n 3 highpass 200 vol 1.67 trim 0 " + sampleSecondsQuota);
					um.exec("rm -f " + path + "input.wav");
					File audFile = new File(path + filename);
					String s1;
					nxtcommand.setVerify(true);
					nxtcommand.uploadFile(audFile, filename);
					nxtcommand.setVerify(false);
					um.exec("rm -f " + path + filename);
				}
				um.exec("rm -f " + path + "mediaList.dat");
				// Also delete any other audio sample files Enchanting exported that were not uploaded.
				um.exec("rm -f " + path + "*.wav");
			}
			
		} catch (Exception e) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Exception occurred processing media: " + e.getMessage());
			System.out.println("Got an IOException in uploadMedia(): " + e.getMessage());
		}
	}
	

	private class FirmwareFlasher implements NXJFlashUI {
		NXJFlashUpdate updater = new NXJFlashUpdate(this);
		
		public void message(String msg) {
			progress(msg, -1);
		}

		public void progress(String msg, int percent) {
			narrowcastMsg(FLASH_STATUS_PREFIX + Integer.toString(percent) + " " + msg);   			
			//System.out.printf("%s %3d%%\n", msg, percent);
		}

		public void flash() throws Exception {
		
			boolean simulate = false;
						
			if (simulate) {
				System.out.println("PRETENDING TO FLASH FIRMWARE");
			
				// There are many more messages sent when this happens for real
				// but this is a rough approximation
				message("Building firmware image.");
				message("Attempting to reboot the device.");
				message("Locating device in firmware update mode.");
				for (int i = 0; i < 100; ++i) { progress("Searching", i); }
				message("Found NXT: %%NXT-SAMBA%%");
				message("Opened device in firmware update mode.");
				message("Writing firmware image.");
				for (int i = 0; i < 100; ++i) { progress("Writing", i); }
				message("Verifying firmware.");
				for (int i = 0; i < 100; ++i) { progress("Verifying", i); }
				message("Verified 94208 bytes ok.");
				message("Restarting the device.");
				
			} else {
				System.out.println("FLASHING FIRMWARE");
			
				File vmFile = null;
				File menuFile = null;
				String home = System.getProperty("nxj.home");
				if (home != null) home = System.getenv("NXJ_HOME");
				byte[] memoryImage = updater.createFirmwareImage(vmFile, menuFile, home);
				byte[] fs = null;
			//	if (commandLine.hasOption("f"))
			//		fs = updater.createFilesystemImage();
			
				// reset the NXT and open it in firmware upgrade mode
				NXTInfo nxtinfo = nxtconnector.getNXTInfo();
				nxtconnector.close();
				updater.resetDevice(nxtinfo);
				NXTSamba nxt = updater.openSambaDevice(30000);
				if (nxt != null) {
					updater.updateDevice(nxt, memoryImage, fs, true, true, true);
				}
			}			
		}
	}
}