Object subclass: #EnchantingCodeExporter
	instanceVariableNames: 'dict functions commands indentationLevel additionalIndent sprite codeStream argDepth template functionIndex outputFileBaseName device spriteNumber allSprites currentCustomCommand '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Enchanting'!
!EnchantingCodeExporter commentStamp: '<historical>' prior: 0!
The EnchantingCodeExported is used to take the scripts of a Scratch sprite and export them into computer code.

dict is a dictionary to convert from the English representation of a code tile to whichever language interests us.

functions is a list of enchanting functions to be exported.

indentationLevel is how many tabs to put in front of a line of code.
additionalIndent is indentation read from a template file, and is prepended to the indentation.

sprite points back to the sprite whose code tiles we are exporting.

codeStream is a stream for placing the code sequences we are building on.

When argDepth is greater than zero, we are within an argument block.  (No tabbing should be performed).

template is the lines of the template file.

functionIndex assigns a number to each function/thread.

sprites is a collection of all the sprites that have been exported.

currentCustomCommand is set when the exporter is traversing through a custom block definition. This allows exportVariableBlockMorph to access the custom block's arguments.

!


!EnchantingCodeExporter methodsFor: 'initialization' stamp: 'CWB 8/8/2011 08:57'!
initialize

	super initialize.
	functions _ Dictionary new.
	commands _ Dictionary new.
	indentationLevel _ 0.
	additionalIndent _ ''.
	argDepth _ 0.
	sprite _ nil.
	spriteNumber _ -1.
	functionIndex _ 1.
	device _ ScriptableScratchMorph blockSpecSource.
	allSprites _ OrderedCollection new.
	currentCustomCommand _ nil.
! !

!EnchantingCodeExporter methodsFor: 'initialization' stamp: 'CWB 9/3/2010 07:08'!
outputFileBaseName: aName
	"Sets the base file name.  If the full name if /path/to/foo.script, the base name is foo."
	"This is needed as in Java, the file is named after the class"

	outputFileBaseName _ aName.! !

!EnchantingCodeExporter methodsFor: 'initialization' stamp: 'CWB 8/19/2010 11:38'!
readTemplateFile: aFilename
	"Reads in the template file"

	| f |
	f _ FileStream readOnlyFileNamedOrNil: 
		(ScratchTranslator translationDir fullNameFor: aFilename).
	f ifNil: [
		DialogBoxMorph inform: 'File not found' withDetails: aFilename.
		^ nil].
	template _ f contentsOfEntireFile lines.
! !


!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/28/2010 09:54'!
codeStream
	"Gets the stream that code is being stored on"

	^ codeStream! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/28/2010 09:55'!
codeStream: aStream
"Sets the stream to store code on.  If nil, creates a new code stream"

	aStream ifNil:
		[ codeStream _ WriteStream on: (String new: 5000) ]
	ifNotNil:
		[ codeStream _ aStream ]! !

!EnchantingCodeExporter methodsFor: 'accessing' stamp: 'CWB 7/21/2010 13:28'!
sprite
	"Grants access to the sprite being exported"

	^ sprite
! !


!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 10:20'!
createUniqueFunctionNameFrom: baseName
	"Ensures that each function has a unique name"

	| appendix name |

	(functions includesKey: baseName) ifFalse: [^ baseName].

	"Keep trying the function name with a new number on the end"
	"until we come up with one that is not in use."

	appendix _ 2.

	[ true ] whileTrue:
	[
		name _ baseName, appendix asString.
		(functions includesKey: name) ifFalse: [^ name].
		appendix _ appendix + 1.
	] 
	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 9/28/2011 20:53'!
customExportColor: aColor inBlock: aBlock
"We are exporting a color.  Its meaning is specific to whichever peripheral is using the block."

	| peripheralArg peripheral |

	peripheralArg _ aBlock findA: ChoiceArgMorphForPeripheral.
	peripheral _ peripheralArg peripheral.
	peripheral ifNil: [ 
		^ self error: 'Unable to export custom color -- unknown peripheral.'
	].

	peripheral customExportColor: aColor onStream: self codeStream.
	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 9/8/2011 16:31'!
exportArgument: aMorph havingSpec: aSpec
	"Takes the given argument to another block, translates if necessary, and outputs the result"

	| debugging |

	debugging _ false.
	debugging ifTrue: [ codeStream nextPutAll: '[', aMorph class asString, ' ', aSpec, ']' ].

	argDepth _ argDepth + 1.

	"A number, a string, a dropdown -- anything except a code block will be an ArgMorph"
	(aMorph isKindOf: ArgMorph) ifTrue: [
		aMorph exportArgWith: self havingSpec: aSpec.
	].

	"Must be a code block"
	(aMorph isKindOf: BlockMorph) ifTrue: [
		aMorph exportCodeWith: self.
	].

	argDepth _ argDepth - 1.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 9/30/2011 22:14'!
exportBlock: aBlock
	"Takes the given block, translates it to code, and outputs the result"

	| englishSpec codeSpec argMapping argIndex codeSpecTempStream |

	self exportCommentAttachedTo: aBlock.
	
	aBlock isHatBlock
		ifTrue:	[ 
			englishSpec _ aBlock englishCodeSpec ]
		ifFalse:	[ 
			englishSpec _ self englishSpecForSelector: aBlock selector.

			"special case for predicates (boolean reporters)"
			"(((englishSpec = 'report %s') 
			and: [currentCustomCommand notNil]) 
			and: [currentCustomCommand type = #boolean]) ifTrue: 
			[
				englishSpec _ 'report %b'.
			] - commented out by JM - all reporters just return EObjects now"
		].


	englishSpec ifNil: [ 
		codeStream nextPutAll: 
			'->NO ENGLISH FOR ' , 
			aBlock class asString asUppercase , 
			' with selector ', aBlock selector asString,
			'<-'. 
		^ nil 
	].

	"Get the code spec"
	(aBlock isCustomBlock)
		ifTrue: 	[ codeSpec _ self codeSpecFromCustomBlock: aBlock. ]
		ifFalse: [ 
			(aBlock isKindOf: MultiArgCommandBlockMorph) | (aBlock isKindOf: MultiArgReporterBlockMorph)
			ifTrue: [ codeSpec _ self codeSpecFromMultiArgBlock: aBlock ]
			ifFalse: [ codeSpec _ self codeSpecFrom: englishSpec ]
		].
		

	"Process any meta commands within the specification."
	"(This makes nested for loops possible.)"
	codeSpecTempStream _ WriteStream on: (String new: 120).

	self processTemplateLine: codeSpec 
		ontoStream: codeSpecTempStream 
		withCommands: [ :command :prevToken |
			command = 'DEPTH' ifTrue: [ codeSpecTempStream nextPutAll: indentationLevel asString ].
		].	

	codeSpec _ codeSpecTempStream contents.

	argMapping _ aBlock class argPermutationForSpec: englishSpec withTranslation: codeSpec.

	self indent.

	argIndex _ 1.
	(CommandBlockMorph parseCommandSpec: codeSpec removeSpaces: false ) do: 
		[ :token |
			(CommandBlockMorph isArgSpec: token) 
				ifTrue: [ 
					| arg |
					"Look up the proper argument"
					(aBlock isCustomBlock) | ( aBlock isKindOf: MultiArgReporterBlockMorph) | ( aBlock isKindOf: MultiArgCommandBlockMorph)
						ifTrue: [ arg _ aBlock argumentAt: argIndex]
						ifFalse: [ arg _ aBlock argumentAt: (argMapping at: argIndex)].

					self exportArgument: arg havingSpec: token.

					argIndex _ argIndex + 1.
				] 
				ifFalse: [ codeStream nextPutAll: token. ].
		].

	(argDepth == 0) ifTrue: [ codeStream cr ].! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/8/2011 09:08'!
exportCommand: aCommand usingName: commandName templateSection: aSection
	"Emits a command with a given name."

	| firstblock type |

	"Keep track of the custom command definition so its arguments can be accesssed."
	currentCustomCommand _ aCommand.

	"Create a new codestream"
	self codeStream: nil.

	"Custom blocks don't appear to retain comments. - CWB"
	"self exportCommentAttachedTo: aCommand."

	firstblock _ aCommand body.
	type _ aCommand type.

	self processTemplateSection: aSection ontoStream: codeStream withCommands: 
	[ :command :prevToken |
		command = 'COMMANDNAME' ifTrue: [ codeStream nextPutAll: commandName ].
		command = 'COMMANDBODY' ifTrue: 
		[
			self usingIndent: prevToken 
				on: codeStream 
				execute: [ 
					firstblock 
						ifNil: [
							self outputTranslatedLineOf: 'empty nested block'.
						]
						ifNotNil: [
							firstblock exportCodeWith: self.
						].
				].
		].	
		(command = 'COMMANDRETURN' and: [type = #any or: [type = #boolean]] ) ifTrue: [
			self exportBlock: aCommand answerBlock.
		].

	].

	currentCustomCommand _ nil.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:54'!
exportCommentAttachedTo: aBlock
"Emits the comment attached to this block (if any)"

	| commentMorph commentText |

	'The inputLines are an array with a new line at every soft linebreak.'.
	'It also has hard line breaks wherever return was pressed.'.
	'outputLines are the same thing, without the line breaks' .

	commentMorph _ aBlock commentMorph.
	commentMorph ifNil: [ ^ nil ].

	commentText _ commentMorph commentMorph contents.

	self putComment: commentText on: codeStream maxLineLength: 60.


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 5/27/2011 22:13'!
exportGreenFlagScript: aBlock
	"Emits code for a script triggered by startup"

	self exportScript: aBlock 
		usingName: 'WhenGreenFlagClicked'
		templateSection: 'BEGIN WHEN GREEN FLAG CLICKED TEMPLATE'
		type: #greenFlag
		trigger: nil.

	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/20/2011 16:56'!
exportListBlockMorph: aMorph
"Exports a list block morph -- trickier than you might think."
"NOTE: If you make changes here, you may need to change exportVariableBlockMorph, too"

	| englishSpec codeSpec argIndex |

	"See if this variable is actually a command parameter"
	currentCustomCommand ifNotNil: [ argIndex _ currentCustomCommand indexForParameterName: (aMorph commandSpec)].
	
	argIndex ifNil: [ englishSpec _ '%L' ] ifNotNil: [ englishSpec _ '%z' ].
	codeSpec _ self codeSpecFrom: englishSpec.

	(CommandBlockMorph parseCommandSpec: codeSpec removeSpaces: false ) do: 
		[ :token |
			(CommandBlockMorph isArgSpec: token) 
				ifTrue: [ 
					"Process as either a command arg or a variable "
					argIndex 
						ifNotNil: [
							codeStream nextPutAll: (currentCustomCommand indexForParameterName: (aMorph commandSpec)) asString.
						]
						ifNil: [ codeStream nextPutAll: (sprite shortNameFor: (aMorph commandSpec)) ].
				]
				ifFalse: [ codeStream nextPutAll: token. ].
		].
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 8/26/2011 10:10'!
exportLiteralValue: value asStringOn: aStream
 	"Exports a literal value that should be non-numeric."
 	"Note that strings that contain quotes have them escaped"
 	"(Helper function for exportLiteralValue:on:)"

 	(value isKindOf: String)
  		ifTrue: [aStream nextPutAll: '"', (value copyReplaceAll: '"' with: '\"'), '"']
  		ifFalse: [aStream nextPutAll: value asString ]
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/20/2011 15:47'!
exportLiteralValue: value on: aStream
	"Exports a literal value (such as might be found in a 'set X to hello' or 'change Y by 1' block)"

	"Assumption -- languages being exported to prefer numbers as numbers (instead of strings)
	when possible, and in the format XXX.XXXf -- with a decimal point and a trailing 'f'"

	| numeric |

	"The empty string evalutes as the number 0!!  We don't want that."
	('' asUTF8 = value) ifTrue: [ ^ self exportLiteralValue: value asStringOn: aStream ].

	[ 	numeric _ value asNumber.
		aStream nextPutAll: numeric asString.
		(numeric asString includes: $.) ifFalse: [ aStream nextPutAll: '.0' ].
		aStream nextPutAll: 'f'.
	] ifError: [ self exportLiteralValue: value asStringOn: aStream ].! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 9/7/2011 10:49'!
exportNestedBlock: aNestedBlock
	"Outputs a compound statement -- a series of statements that act rather like one statement."

	self outputTranslatedLineOf: 'nested block start'.
	indentationLevel _ indentationLevel + 1.

	aNestedBlock 
		ifNil: [
			self outputTranslatedLineOf: 'empty nested block'.
		]
		ifNotNil: [
			aNestedBlock exportCodeWith: self.
		].
 
	indentationLevel _ indentationLevel - 1.
	self outputTranslatedLineOf: 'nested block end'.
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 5/27/2011 22:13'!
exportScript: aBlock usingName: baseName templateSection: aSection type: aType trigger: aTrigger
	"Emits a function with a given base name."

	| functionName |

	functionName _ self createUniqueFunctionNameFrom: baseName.

	"Create a new code stream"
	self codeStream: nil.

	self exportCommentAttachedTo: aBlock.

	self processTemplateSection: aSection ontoStream: codeStream withCommands: 
	[ :command :prevToken |
		command = 'FUNCTIONNAME' ifTrue: [ codeStream nextPutAll: functionName ].
		command = 'FUNCTIONTRIGGER' ifTrue: [ codeStream nextPutAll: aTrigger ].
		command = 'FUNCTIONBODY' ifTrue: 
		[
			self usingIndent: prevToken 
				on: codeStream 
				execute: [ 
					aBlock nextBlock 
						ifNil: [
							self outputTranslatedLineOf: 'empty nested block'.
						]
						ifNotNil: [
							aBlock nextBlock exportCodeWith: self.
						].
				]
		].
	].
		
	"Store everything into a function structure"
	functions at: functionName put: 
		(EnchantingFunction 
			newWithSource: codeStream contents
			type: aType
			index: functionIndex).
	functionIndex _ functionIndex + 1.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 10/3/2011 11:17'!
exportVariableBlockMorph: aMorph
"Exports a variable block morph -- trickier than you might think."

	| englishSpec codeSpec argIndex |

	"See if this variable is actually a command parameter"
	currentCustomCommand ifNotNil: [ argIndex _ currentCustomCommand indexForParameterName: (aMorph commandSpec) ].
	
	argIndex ifNil: [ englishSpec _ '%v' ] ifNotNil: [ englishSpec _ '%z' ].
	codeSpec _ self codeSpecFrom: englishSpec.

	(CommandBlockMorph parseCommandSpec: codeSpec removeSpaces: false ) do: 
		[ :token |
			(CommandBlockMorph isArgSpec: token) 
				ifTrue: [ 
					"Process as either a command arg or a variable "
					argIndex 
						ifNotNil: [
							codeStream nextPutAll: ((currentCustomCommand indexForParameterName: (aMorph commandSpec)) - 1) asString.
						]
						ifNil: [ codeStream nextPutAll: (sprite shortNameFor: (aMorph variable)) ].
				]
				ifFalse: [ codeStream nextPutAll: token. ].
		].
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 5/27/2011 22:14'!
exportWhenIReceiveScript: aBlock trigger: aTrigger
	"Emits code for a script triggered when I specific message is broadcast"

	self exportScript: aBlock 
		usingName: 'WhenIReceive'
		templateSection: 'BEGIN WHEN I RECEIVE TEMPLATE'
		type: #broadcast
		trigger: aTrigger.

	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:48'!
indent

	"Outputs indentation for the current line of code"

	self indentOn: codeStream.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 16:01'!
indentOn: aStream

	"Outputs indentation for the current line of code"

	| indentText |

	indentText _ self codeSpecFrom: 'indent'.

	argDepth == 0 ifTrue: 
	[ 
		aStream nextPutAll: additionalIndent.
		indentationLevel timesRepeat: [ aStream nextPutAll: indentText ].
	]! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 8/9/2011 15:55'!
printEachCommandOn: aStream usingTemplateSection: aSectionName ofType: aType
	"The template can call for a section of code to be invoked for each custom command.  This function makes it happen."
	| nameToSpec addToMap stage spec functionName |

	"Populate a map that maps all function names onto the corresponding block."
	"Note that sprite1 can have a block called 'foo' that is different from sprite2's block called 'foo'"
	"The function names are unique keys in the dictionary; the blocks' specs are not always unique."

	nameToSpec _ Dictionary new.

	addToMap _ [:aSprite :aBlock |
		spec _ aBlock userSpec.
		functionName _ aSprite customBlockFunctionNameFor: spec.
		nameToSpec at: functionName put: aBlock.
	].

	stage _ allSprites first.
	(stage sortedCustomBlocks select: [:b | b isGlobal]; select: [ :b | b type = aType ]) 
		do: [:block | addToMap value: stage value: block ].

	allSprites do: [ :aSprite |
		(aSprite sortedCustomBlocks reject: [:b | b isGlobal]; select: [ :b | b type = aType ]) 
			do: [:block | addToMap value: aSprite value: block ]
	].
	

	"Now, emit code for each function"

	nameToSpec keysAndValuesDo: [ :fnName :block |
		self exportCommand: block usingName: fnName templateSection: aSectionName.
		aStream nextPutAll: codeStream contents.
	].

	! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/21/2010 12:04'!
printEachFunctionOn: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each function/thread.  This function makes it happen."

	functions keysDo: [:key |
		| func |

		func _ functions at: key.		

		self processTemplateSection: aSectionName ontoStream: aStream withCommands: 
		[:command :prevToken |
			command = 'FUNCTIONNAME' ifTrue: [ aStream nextPutAll: key ].
			command = 'FUNCTIONINDEX' ifTrue: [ aStream nextPutAll: func index asString ].
		].
	].

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/11/2011 13:21'!
printEachListOn: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each list.  This function makes it happen."

	allSprites do: [:s |
		s listVarNames asSortedCollection do: [:l |
			self printList: l of: s on: aStream usingTemplateSection: aSectionName.
		]
	]	
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 5/31/2011 22:09'!
printEachRemoteSensorOn: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each remote sensor (ie. a global variable from Scratch).  This function makes it happen."

	| stage server |

	stage _ allSprites first.
	server _ stage scratchServer.

	server ifNotNil: [
		server sensorNames do: [:s |
			| value |
			value _ server sensorValueFor: s.
			self printRemoteSensor: s value: value on: aStream usingTemplateSection: aSectionName.
		].
	].

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 8/10/2011 11:48'!
printEachVariableOn: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each variable.  This function makes it happen."

	allSprites do: [:s |
		s varNames asSortedCollection do: [:v |
			self printVariable: v of: s on: aStream usingTemplateSection: aSectionName.
		]
	]	
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/11/2011 13:29'!
printList: aListName of: aSprite on: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each list.  This function makes it happen."

	| value shortName fullName |

	fullName _ aListName.
	shortName _ aSprite shortNameFor: fullName.
	"value _ (aSprite getVar: aListName) asString."
	value _ ''.

	self processTemplateSection: aSectionName ontoStream: aStream withCommands: 
	[:command :prevToken |
		command = 'LISTNAME'  		ifTrue: [ aStream nextPutAll: shortName ].
		command = 'LISTFULLNAME'	ifTrue: [ aStream nextPutAll: fullName ].
		command = 'LISTVALUE' 		ifTrue: [ aStream nextPutAll: value ].
	].

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/11/2011 13:07'!
printOn: aStream
	"Prints the exported code onto the given stream"

	"Bail out if we don't have all the information we need"
	template ifNil: [aStream nextPutAll: 'Could not read template file.'].
	dict ifNil: [aStream nextPutAll: 'Could not read dictionary file.'].
	device ifNil: [aStream nextPutAll: 'No enchanting device is configured.'].
	((template isNil) or: [dict isNil]) ifTrue: [^nil].

	additionalIndent _ ''.

	self processTemplateSection: 'BEGIN FILE TEMPLATE' ontoStream: aStream withCommands: 
	[:command :prevToken |
		command = 'ENCHANTING HEADER' ifTrue: [ self putHeaderOn: aStream ].
		command = 'IMPORTS' ifTrue: [ "to do" ].
		command = 'BASEFILENAME' ifTrue: [ aStream nextPutAll: outputFileBaseName ].
		command = 'FUNCTIONS' ifTrue: 
		[
			self usingIndent: prevToken on: aStream execute: [ self putFunctionsOn: aStream ].
		].
		command = 'INITIALIZE PERIPHERALS' ifTrue: 
		[
			self usingIndent: prevToken 
				on: aStream 
				execute: [ device emitInitializationCodeOn: aStream for: self ].
		].
		('USE "*" FOR EACH FUNCTION' match: command) ifTrue:
		[
			| sectionName |
			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachFunctionOn: aStream usingTemplateSection: sectionName. ] 
		].
		('USE "*" FOR EACH VARIABLE' match: command) ifTrue:
		[
			| sectionName z |

			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachVariableOn: aStream usingTemplateSection: sectionName. ]
		].
		('USE "*" FOR EACH LIST' match: command) ifTrue:
		[
			| sectionName z |

			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachListOn: aStream usingTemplateSection: sectionName. ]
		].
		('USE "*" FOR EACH REMOTE SENSOR' match: command) ifTrue:
		[
			| sectionName z |

			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachRemoteSensorOn: aStream usingTemplateSection: sectionName. ]
		].
		('USE "*" FOR EACH COMMAND' match: command) ifTrue:
		[
			| sectionName z |

			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachCommandOn: aStream usingTemplateSection: sectionName ofType: #none. ]
		].
		('USE "*" FOR EACH REPORTER' match: command) ifTrue:
		[
			| sectionName z |

			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachCommandOn: aStream usingTemplateSection: sectionName ofType: #any. ]
		].
		('USE "*" FOR EACH PREDICATE' match: command) ifTrue:
		[
			| sectionName z |

			sectionName _ (command findTokens: '"') second.	
			self usingIndent: prevToken 
				on: aStream
				execute: [ self printEachCommandOn: aStream usingTemplateSection: sectionName ofType: #boolean. ]
		].
	].
			

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 5/31/2011 22:11'!
printRemoteSensor: aName value: aValue on: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each variable.  This function makes it happen."

	self processTemplateSection: aSectionName ontoStream: aStream withCommands: 
	[:command :prevToken |
		command = 'REMOTESENSORNAME'  	ifTrue: [ aStream nextPutAll: aName ].
		command = 'REMOTESENSORVALUE' 	ifTrue: [ aStream nextPutAll: (aValue asString) ].
	].

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 9/20/2011 21:34'!
printVariable: aVarName of: aSprite on: aStream usingTemplateSection: aSectionName
	"The template can call for a section of code to be invoked for each variable.  This function makes it happen."

	| value shortName fullName |

	fullName _ aVarName.
	shortName _ aSprite shortNameFor: fullName.
	value _ (aSprite getVar: aVarName) asString.

	self processTemplateSection: aSectionName ontoStream: aStream withCommands: 
	[:command :prevToken |
		command = 'VARIABLENAME'  		ifTrue: [ aStream nextPutAll: shortName ].
		command = 'VARIABLEFULLNAME'	ifTrue: [ aStream nextPutAll: fullName ].
		command = 'VARIABLEVALUE' 		ifTrue: [ self exportLiteralValue: value on: aStream ].
	].

	


! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 9/3/2010 07:12'!
processScript: aScript

	"Takes a script (a hat block that starts a list of blocks) and processes it for output from the exporter"

	"The first thing the hat block will do is create a new function, complete with code stream"

	aScript exportCodeWith: self.

! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:55'!
putComment: aComment on: aStream maxLineLength: maxLength

	"Takes a comment which may have multiple lines and puts it onto a stream."

	(aComment withNoLineLongerThan: maxLength) lines do: 
	[:line |
		self putSingleLineComment: line on: aStream
	].! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:56'!
putFunctionsOn: aStream

	functions do: [:f | aStream nextPutAll: f source; cr. ]
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 12:26'!
putHeaderOn: aStream
	"Emits information about enchanting, and places it as a comment on the stream"

	self putComment: 'Exported from Enchanting' on: aStream maxLineLength: 60.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/19/2010 11:49'!
putSingleLineComment: aComment on: aStream 
	"Outputs a comment onto a stream"

	self indentOn: aStream.
	aStream nextPutAll: (self translationOf: 'single-line comment start' ifAbsent: ['']).
	aStream nextPutAll: aComment.
	aStream nextPutAll: (self translationOf: 'single-line comment end' ifAbsent: ['']).
	aStream cr.! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'JM 8/9/2011 15:57'!
sprite: aSprite
	"Sets the sprite to be operated on."

	| prefix |
	
	(sprite _ aSprite) ifNotNil: [	
		spriteNumber _ spriteNumber + 1.
		Transcript show: 'sprite #', spriteNumber asString, ' (', sprite objName, ')'; cr.
		(spriteNumber = 0) 
			ifTrue: [ 
				prefix _ 'stage_'.
				sprite createFunctionNamesForGlobalCustomBlocksPrefixedBy: 'global_'.
			]
			ifFalse: [ prefix _ 'sprite', spriteNumber asString, '_' ].
		sprite createShortNamesForVariablesPrefixedBy: prefix.
		"sprite createMapForSoundMediaPrefixedBy: prefix." "audio export disabled for now"
		sprite createFunctionNamesForCustomBlocksPrefixedBy: prefix.
		allSprites add: sprite.
	]
! !

!EnchantingCodeExporter methodsFor: 'exporting' stamp: 'CWB 8/21/2010 12:36'!
usingIndent: aNewIndent on: aStream execute: aBlock 
	"Pushes a new indentation level, executes the code, pops the indentation level.  Useful for formatting replaced code within templates"

	| prevIndent |

	prevIndent _ additionalIndent.
	additionalIndent _ aNewIndent.
	
	"The indent will already be on the stream; remove it so it is not there twice"
	aStream skip: (aNewIndent size) * -1.
	"aStream nextPutAll: 'REMOVED ->', aNewIndent, '<- (', aNewIndent size asString, ')'."

	aBlock value.

	"remove the carriage return following the command to insert code from a template"
	aStream skip: -1.

	additionalIndent _ prevIndent.	

! !


!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 9/30/2011 21:58'!
processTemplateLine: aLine ontoStream: aStream withCommands: commandBlock
	"Goes through a line with template metacharacters in it and emits text or acts on meta commands."
	"See notes and examples at bottom."

	| metaChar isCommand prevToken tokens |

	metaChar _ self templateMetaCharacter.
	isCommand _ false.
	prevToken _ ''.

	"self indentOn: aStream."  "Seems to have been responsible for double-indents."

	tokens _ aLine findTokens: metaChar keep: metaChar.
	tokens do: [:token |
		token = metaChar 
			ifTrue: [ isCommand _ isCommand not ]
			ifFalse: [
				isCommand 
					ifTrue: [ 
						"do normal processing only if there is not a meta-command present"
						(token = 'METACHAR') ifTrue: 
						[
							aStream nextPutAll: (metaChar asString).
						] ifFalse: [
							commandBlock value: token value: prevToken.
						].
					]
					ifFalse: [
						prevToken _ token.
						aStream nextPutAll: token.
					]
			] 
	].
	
	"(The token count is used to inhibit new lines in special circumstances)"
	^ tokens size.

"This method allows you to take a string and put commands in it along with text to
be exported as-is.  The commands typically write something to the same stream,
but substitute a value in for it."

"Example of using this method:"
"
	e _ EnchantingCodeExporter newWithSprite: nil exporter: 'lejos'.
	e processTemplateLine: 'int i@DEPTH@ = @DEFAULT@' 
	  ontoStream: Transcript 
       withCommands: 
		[:command :prevToken | 
			command = 'DEPTH' ifTrue: [ Transcript nextPutAll: '47' ]. 
			command = 'DEFAULT' ifTrue: [ Transcript nextPutAll: '0' ]. 
		]
Transcript flush.
"

"Output is: int i47 = 0"

	! !

!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 10/19/2010 20:41'!
processTemplateSection: aSectionName ontoStream: aStream withCommands: commandBlock
	"Goes through the template (a collection of lines), emits text or acts on meta commands."

	| sectionTemplate |

	sectionTemplate _ self templateSection: aSectionName.

	sectionTemplate do: [:line |
		| tokenCount |
		tokenCount _ self processTemplateLine: line 
			ontoStream: aStream 
			withCommands: commandBlock.

		"We use tokenCount to inhibit placing a newline at the end of a line"
		"that was only a single command and nothing else."
		"The situation is the only one that will yield 3 tokens -- meta, command, and meta"
		tokenCount ~= 3 ifTrue: [aStream cr].		
	]
! !

!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 10/18/2010 21:14'!
templateMetaCharacter
	"Returns the template meta-character (as a string)."
	"(This character delimits portions of the template that will be filled in based on what the user has created in enchanting.)"

	^ self translationOf: 'template meta sequence' ifAbsent: ['@'].! !

!EnchantingCodeExporter methodsFor: 'templates' stamp: 'CWB 8/19/2010 15:15'!
templateSection: aTitle
	"Returns a portion of the template bounded by aTitle and 'END TEMPLATE', both within metacharacters"

	| metaChar startLine endLine startIndex endIndex |

	metaChar _ self templateMetaCharacter.
	startLine _ metaChar, aTitle, metaChar.
	endLine _ metaChar, 'END TEMPLATE', metaChar.

	startIndex _ template indexOf: startLine.
	startIndex = 0 ifTrue: [^nil].

	endIndex _ template indexOf: endLine startingAt: startIndex ifAbsent: [ 0 ].
	endIndex = 0 ifTrue: [^nil].

	^ template copyFrom: startIndex + 1 to: endIndex - 1. ! !


!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 12/20/2010 20:30'!
codeSpecFrom: english
	"Returns the specification for the block in the computer language we are exporting to"

	^ dict at: (english asUTF8) ifAbsent: [ '>>?', english, '?<<' ].! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'JM 9/7/2011 11:53'!
codeSpecFromCustomBlock: aBlock

	| spec decs length count |

	decs _ aBlock definition declarations.

	"First determine the type of custom block"
	spec _ self sprite customBlockFunctionNameFor: aBlock userSpec.
	spec _ spec, (self codeSpecFrom: (aBlock definition englishSpec)).
	spec _ spec , (self codeSpecFrom: 'call start').

	length _ decs size.
	count _ 1.

	decs size - 1 timesRepeat: [

		spec _ spec , '%z'.

		"Don't place a separator after the final element"
		(count < length) ifTrue: [
			spec _ spec, (self codeSpecFrom: 'argument separator').
		].
		count _ count + 1.
	].

	"Command blocks end differently to reporters and predicates (need a semicolon)"
	aBlock definition type = #none 
		ifTrue: [ spec _ spec, (self codeSpecFrom: 'command end') ]
		ifFalse: [ spec _ spec, (self codeSpecFrom: 'call end') ].
	
	^ spec.! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'JM 9/30/2011 16:54'!
codeSpecFromMultiArgBlock: aBlock

	| count spec numArgs name end argSymbol |

	"if there's no arguments then just return the block's command spec"
	aBlock argMorphs size = 1 ifTrue: [ ^ self codeSpecFrom: aBlock commandSpec ].

	"different command specs for command / reporter call"
	( aBlock selector = #doCallBlockWithArgs) 
		ifTrue: [ name _ 'call %r with'. end _ 'call end' ].

	( aBlock selector = #doCallBlockWithArgList)
		ifTrue: [ name _ 'call %r with list'. end _ 'call with list end' ].

	( aBlock selector = #doRunBlockWithArgs)
		ifTrue: [ name _ 'run %o with'. end _ 'command end' ].

	( aBlock selector = #doRunBlockWithArgList)
		ifTrue: [ name _ 'run %o with list'. end _ 'run with list end' ].

	( aBlock selector = #doForkBlockWithArgs)
		ifTrue: [ name _ 'launch %o with'. end _ 'launch end' ].

	( aBlock selector = #doForkBlockWithArgList)
		ifTrue: [ name _ 'launch %o with list'. end _ 'launch with list end' ].



	( aBlock selector endsWith: 'List' ) 
		ifTrue: [ argSymbol _ '%t' ]
		ifFalse: [ argSymbol _ '%s' ].
		
	spec _ (self codeSpecFrom: name ) , '%A'.

	numArgs _ aBlock argMorphs size - 2.
	count _ 1.

	numArgs timesRepeat: [

		spec _ spec , argSymbol.

		"Don't place a separator after the final element"
		(count < numArgs) ifTrue: [
			spec _ spec, (self codeSpecFrom: 'argument separator').
		].
		count _ count + 1.
	].

	spec _ spec , (self codeSpecFrom: end).

	^ spec.

	
	! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 9/4/2010 21:40'!
englishSpecForSelector: aSelector
	"Returns the specification for the block in the native english"

	| english |

	english _ self sprite blockSpecForSelector: aSelector.	
	english ifNotNil: [ ^english ].

	"If you are here, you've found a block that doesn't get english from a selector."
	"Please add a special case to ScriptableScratchMorph>>blockSpecForSelector"

	self halt.
	^ nil.! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'JM 9/30/2011 11:41'!
exportCommandScript: aBlock

	"exports a 'the script' or 'the block' block, including the short hand grey border versions"

	| name spec length count prevArgDepth |

	((aBlock isKindOf: CommandScriptBlockMorph) | (aBlock isKindOf: CSlotMorph)) ifTrue: [ name _ 'script' ] ifFalse: [ name _ 'block' ].
		
	spec _ ''.
	length _ aBlock declarations size.

	"Export start of script code"
	self outputTranslationOfTerm: 'the ' , name.

	"Add the arguments"
	count _ 1.

	length - 1 timesRepeat: [

		spec _ spec , '%z'.

		"Don't place a separator after the final element"
		(count < length) ifTrue: [
			spec _ spec , (self codeSpecFrom: 'argument separator').
		].
		count _ count + 1.
	].

	codeStream nextPutAll: spec; cr.

	indentationLevel _ indentationLevel + 1.

	"argDepth keeps track of if we are exporting an argument or not, and suppress indentation if we are."
	"Here we pretend that we are not, in fact, in an argument, to make the code export in a pretty fashion."
	prevArgDepth _ argDepth.
	argDepth _ 0.

	"Add the inner blocks"
	currentCustomCommand _ aBlock.

	aBlock nestedBlock 
		ifNil: [ self outputTranslatedLineOf: 'empty nested block'. ]
		ifNotNil: [ aBlock nestedBlock exportCodeWith: self. ].

	currentCustomCommand _ nil.

	indentationLevel _ indentationLevel - 1.

	"Export end of script code"
	self indent; outputTranslationOfTerm: 'the ' , name , ' end'.

	argDepth _ prevArgDepth.

! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/19/2010 11:07'!
importTranslation: filename
	"Reads a translation dictionary from the specified file"
 
	^ dict _ ScratchTranslator importTranslation: filename
	! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/16/2010 21:16'!
outputTranslatedLineOf: aTerm
	"Outputs the translation of a given term (if it exists) as a whole line."

	(dict includesKey: aTerm)
		ifTrue: [
			self indent; outputTranslationOfTerm: aTerm.
			codeStream cr. 
		]! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 8/14/2010 21:58'!
outputTranslationOfTerm: englishTerm
	"Takes an english term (such as appears in a choice argument morph), and translates it onto the code stream"

	codeStream nextPutAll: (self codeSpecFrom: englishTerm).! !

!EnchantingCodeExporter methodsFor: 'translating' stamp: 'CWB 9/2/2011 15:05'!
translationOf: english ifAbsent: block
	"Returns the specification for the block in the computer language we are exporting to"

	^ dict at: english asUTF8 ifAbsent: block! !


!EnchantingCodeExporter methodsFor: 'debugging' stamp: 'CWB 9/4/2010 08:46'!
showVariables
"Debugging function to show what all the variables are.  Please be sure the sprite is set before calling this."

	
	| s g allVars |

	s _ WriteStream on: (String new: 5000).

	s nextPutAll: 'Local Variables'; cr; nextPutAll: '---------'; cr.

	sprite varNames do: [:n |
		s nextPutAll: n, ' = ', (sprite getVar: n) asString; cr].

	g _ sprite ownerThatIsA: ScratchStageMorph.
		
	s cr; nextPutAll: 'Global Variables'; cr; nextPutAll: '---------'; cr.

	g varNames do: [:n |
		s nextPutAll: n, ' = ', (g getVar: n) asString; cr].

	allVars _ sprite varNames, g varNames.
allVars do: [:n | s nextPutAll: n; cr].	

	s cr; nextPutAll: 'Local Lists'; cr; nextPutAll: '---------'; cr.

	sprite listVarNames do: [:ln |
		s nextPutAll: ln; cr.
		(sprite valuesOfList: ln) do: [:item |
			s nextPutAll: '   ', item; cr.
		]
	].
	

	s cr; nextPutAll: 'Global Lists'; cr; nextPutAll: '---------'; cr.

	g listVarNames do: [:ln |
		s nextPutAll: ln; cr.
		(g valuesOfList: ln) do: [:item |
			s nextPutAll: '   ', item; cr.
		]
	].

	(StringHolder new contents: s contents) openLabel: 'Variables'.! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

EnchantingCodeExporter class
	instanceVariableNames: ''!

!EnchantingCodeExporter class methodsFor: 'instance creation' stamp: 'CWB 8/19/2010 11:29'!
newWithSprite: aSprite exporter: baseFileName 
	"Reads in the given file which has instructions on how to translate"

	^ self new
		importTranslation: baseFileName, '.exporter';
		readTemplateFile: baseFileName, '.template';
		sprite: aSprite

! !

!EnchantingCodeExporter class methodsFor: 'instance creation' stamp: 'CWB 8/26/2010 20:00'!
newWithSprite: aSprite exporter: baseFileName outputAs: baseOutputName
	"Reads in the given file which has instructions on how to translate"

	^ self new
		importTranslation: baseFileName, '.exporter';
		readTemplateFile: baseFileName, '.template';
		outputFileBaseName: baseOutputName;
		sprite: aSprite

! !
